
/* 
 *  M_APM  -  primenum.c
 *
 *  Copyright (C) 1999 - 2001   Michael C. Ring
 *
 *  Permission to use, copy, and distribute this software and its
 *  documentation for any purpose with or without fee is hereby granted, 
 *  provided that the above copyright notice appear in all copies and 
 *  that both that copyright notice and this permission notice appear 
 *  in supporting documentation.
 *
 *  Permission to modify the software is granted, but not the right to
 *  distribute the modified code.  Modifications are to be distributed 
 *  as patches to released version.
 *  
 *  This software is provided "as is" without express or implied warranty.
 */

/*
 *      $Id: primenum.c,v 1.4 2001/02/18 16:02:33 mike Exp $
 *
 *      PRIME Number Generator using the MAPM Library
 *
 *	$Log: primenum.c,v $
 *	Revision 1.4  2001/02/18 16:02:33  mike
 *	filter out multiples of 2,3,5, and 7
 *
 *	Revision 1.3  2001/02/15 23:11:13  mike
 *	slightly improved algorithm
 *
 *	Revision 1.2  1999/07/12 02:36:23  mike
 *	added more usage and added more comments
 *
 *	Revision 1.1  1999/07/12 02:31:23  mike
 *	Initial revision
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "m_apm.h"

extern  int     is_number_prime(M_APM);

#define FALSE 0
#define TRUE 1

char	 buffer[2048];

int main(argc, argv)
int argc;  char *argv[];
{
int      ct;
				/* declare the M_APM variables ... */
M_APM    aa_mapm;
M_APM    bb_mapm;
M_APM    cc_mapm;
M_APM    dd_mapm;

if (argc < 2)
  {
   fprintf(stdout,"Usage: primenum number\t\t\t\t[Version 1.1]\n");
   fprintf(stdout,
      "       find the first 10 prime numbers starting with \'number\'\n");

   exit(4);
  }
				/* now initialize the M_APM variables ... */
aa_mapm = m_apm_init();
bb_mapm = m_apm_init();
cc_mapm = m_apm_init();
dd_mapm = m_apm_init();

m_apm_set_string(dd_mapm, argv[1]);

if (m_apm_compare(dd_mapm, MM_Three) == -1)
  {
   m_apm_copy(dd_mapm, MM_Three);
  }

m_apm_integer_divide(aa_mapm, dd_mapm, MM_Two);
m_apm_multiply(bb_mapm, MM_Two, aa_mapm);
m_apm_add(aa_mapm, MM_One, bb_mapm);

ct = 0;

while (TRUE)
  {
   if (is_number_prime(aa_mapm))
     {
      m_apm_to_integer_string(buffer, aa_mapm);
      fprintf(stdout,"%s\n",buffer);

      if (++ct == 10)
        break;
     }

   m_apm_add(cc_mapm, MM_Two, aa_mapm);
   m_apm_copy(aa_mapm, cc_mapm);
  }

m_apm_free(aa_mapm);
m_apm_free(bb_mapm);
m_apm_free(cc_mapm);
m_apm_free(dd_mapm);

exit(0);
}
/****************************************************************************/
/*
 *      functions returns TRUE if the M_APM input number is prime
 *                        FALSE if it is not
 */
int     is_number_prime(input)
M_APM   input;
{
static  int     nflg = 1;
static  M_APM   limit, digit, quot, rem, tmp0;
int     ii, ret, index;
char    sbuf[32];

/*
 *      this increment table will filter out all numbers
 *      that are multiples of 2,3,5 and 7.
 *
 *      for reference:
 *
 *      table size of 2 to filter multiples of 2 and 3 
 *      table size of 8 to filter multiples of 2, 3 and 5
 *      table size of 480 to filter multiples of 2,3,5,7, and 11
 */

static  char  incr_table[48] = {
        2, 4, 2, 4, 6, 2, 6, 4, 2, 4, 6, 6, 2, 6, 4, 2,
        6, 4, 6, 8, 4, 2, 4, 2, 4, 8, 6, 4, 6, 2, 4, 6,
        2, 6, 6, 4, 2, 4, 6, 2, 6, 4, 2, 4, 2, 10, 2, 10 };
   

if (nflg)
  {
   nflg  = 0;
   limit = m_apm_init();
   digit = m_apm_init();
   quot  = m_apm_init();
   rem   = m_apm_init();
   tmp0  = m_apm_init();
  }

/* 
 *  since the real algorithm starts at 11 (to syncronize
 *  with the increment table), we will cheat for numbers < 10.
 */

if (m_apm_compare(input, MM_Ten) <= 0)
  {
   m_apm_to_integer_string(sbuf, input);
   ii = atoi(sbuf);

   if (ii == 2 || ii == 3 || ii == 5 || ii == 7)
     return(TRUE);
   else
     return(FALSE);
  }

ret   = FALSE;
index = 0;

m_apm_integer_div_rem(quot, rem, input, MM_Three);
if (m_apm_sign(rem) == 0)               /* remainder == 0 */
  return(ret);

m_apm_integer_div_rem(quot, rem, input, MM_Five);
if (m_apm_sign(rem) == 0)
  return(ret);

m_apm_set_long(digit, 7L);
m_apm_integer_div_rem(quot, rem, input, digit);
if (m_apm_sign(rem) == 0)
  return(ret);

ii = m_apm_exponent(input) + 16;

m_apm_sqrt(tmp0, ii, input);
m_apm_add(limit, MM_Two, tmp0);
   
m_apm_set_long(digit, 11L);              /* now start at '11' to check */
   
while (TRUE)
  {
   if (m_apm_compare(digit, limit) >= 0)
     {
      ret = TRUE;
      break;
     }
   
   m_apm_integer_div_rem(quot, rem, input, digit);
   
   if (m_apm_sign(rem) == 0)         /* remainder == 0 */
     break;
   
   m_apm_set_long(rem, (long)incr_table[index]);
   m_apm_add(tmp0, digit, rem);
   m_apm_copy(digit, tmp0);

   if (++index == 48)
     index = 0;
  }

return(ret);
}
/****************************************************************************/
